/* 

aa.cpp

Jason Sanders, jason.sanders@physics.ox.ac.uk, adapted from code by Andy Eyre

Contains functions which define Stackel Potential, the Coordinate System and for printing

*/
//=================================================================================================

#include <cmath>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <stdlib.h>
#include "GFn.h"
#include "aa.h"
#include "PotMN.h"
#include "fitpot.h"
#include "utils.h"

#define PI 3.1415926535897932384626433832795

#define TINY 1e-15
#define EPS 3.0e-8
#define SMALL 1e-5
#define SIGN(A) ((A)>0.?(1.0):(-1.0))
#define MIN(A,B) ((A)<(B)?(A):(B))
#define MAX(A,B) ((A)>(B)?(A):(B))
#define SIGN2(A,B) ((B)>=0 ?fabs(A):-fabs(A))

//=================================================================================================
// Globals //

double IntGL[3];
double taubargl,Deltagl;
extern double G(double tau);
extern double GPrime(double tau);
double ACCURACY=1.;

//=================================================================================================
// Integrands //

double ptau2(double tau,double *Int){
	//p^2(tau) using local integrals
	return (Int[0]-Int[1]/(tau+Alpha)-Int[2]/(tau+Gamma)+G(tau))/(2.0*(tau+Alpha));
	}
	
double ptau2(double tau){
	//p^2(tau) using global integrals
	return (IntGL[0]-IntGL[1]/(tau+Alpha)-IntGL[2]/(tau+Gamma)+G(tau))/(2.0*(tau+Alpha));
	}

double ptilde(double lam, double lower, double upper){
	return sqrt(ptau2(lam)/((lam-lower)*(upper-lam)));
	}
	
double ptau2ROOT(double tau){
	//p^2(tau)*2.0*(tau+Alpha) - used for finding roots
	return (IntGL[0]-IntGL[1]/(tau+Alpha)-IntGL[2]/(tau+Gamma)+G(tau));
	}
	
double Jint(double theta){
	// need to set taubargl and Deltagl
	double tau=taubargl+Deltagl*sin(theta);
	return sqrt(MAX(0.,ptau2(tau)))*cos(theta);
}

double dJdEint(double theta){
	// need to set taubargl and Deltagl
	double tau=taubargl+Deltagl*sin(theta);
	return 0.25*cos(theta)/((tau+Alpha)*sqrt(ptau2(tau)));
}

double dJdI2int(double theta){
	// need to set taubargl and Deltagl
	double tau=taubargl+Deltagl*sin(theta);
	return -0.25*cos(theta)/((tau+Alpha)*(tau+Alpha)*sqrt(ptau2(tau)));
}

double dJdI3int(double theta){
	// need to set taubargl and Deltagl
	double tau=taubargl+Deltagl*sin(theta);
	return -0.25*cos(theta)/((tau+Alpha)*(tau+Gamma)*sqrt(ptau2(tau)));
}

//=================================================================================================
// Coordinate System Class Functions

CoordSys::CoordSys(double A, double G){
// This constructor sets the parameters alpha and gamma which define the prolate spheroidal coordinate system
// These values are then used to fill the common structure which tells the Fortran code which coordinate system we are using
	Alpha = A;
	Gamma = G;
		}

void CoordSys::xyz2tau(double x[], double tau[]){
  // Calculates the prolate spheroidal coordinates tau = (lambda,phi,nu) given cartesian coordinates x = (x,y,z)
    
  double R2 = x[0]*x[0]+x[1]*x[1];

  double b = Alpha+Gamma-R2-x[2]*x[2];
  double c = Alpha*Gamma-Gamma*R2-Alpha*x[2]*x[2];

  tau[0] = 0.5*(-b+sqrt(b*b-4.0*c)); // lambda
  tau[2] = 0.5*(-b-sqrt(b*b-4.0*c)); // nu
  tau[1] = atan2(x[1],x[0]); // in radians
  return;
      }

void CoordSys::Rz2tau(double x[], double tau[]){
  // Calculates the prolate spheroidal coordinates tau = (lambda,phi,nu) given cartesian coordinates x = (x,y,z)
    
  double R2 = x[0]*x[0];

  double b = Alpha+Gamma-R2-x[1]*x[1];
  double c = Alpha*Gamma-Gamma*R2-Alpha*x[1]*x[1];

  tau[0] = 0.5*(-b+sqrt(b*b-4.0*c)); // lambda
  tau[1] = 0.5*(-b-sqrt(b*b-4.0*c)); // nu

  return;
      }

void CoordSys::xv2taudot(double x[3], double v[3], double taudot[3]){
  /// Calculates taudot given cartesian coordinates x and velocities v
 	
  	double R2 = x[0]*x[0]+x[1]*x[1];

  	double b = Alpha+Gamma-R2-x[2]*x[2];
  	double c = Alpha*Gamma-Gamma*R2-Alpha*x[2]*x[2];

  	double bdot = -2.0*(x[0]*v[0]+x[1]*v[1]+x[2]*v[2]);
  	double cdot = -2.0*(Gamma*(x[0]*v[0]+x[1]*v[1])+Alpha*x[2]*v[2]);
  	
	taudot[0] = 0.5*(-bdot+(b*bdot-2.0*cdot)/sqrt(b*b-4.0*c));
	taudot[2] = 0.5*(-bdot-(b*bdot-2.0*cdot)/sqrt(b*b-4.0*c));
	
	taudot[1] = (x[0]*v[1]-v[0]*x[1])/R2;
	
	return;
	
      }
      
void CoordSys::tau2xyz(double tau[3], double taudot[3], double x[3], double v[3]){
// Calculates the Cartesian coordinates (x,v) given the tau coordinates

	double R2 = (tau[0]+Alpha)*(tau[2]+Alpha)/(Alpha-Gamma);
	double z2 = (tau[0]+Gamma)*(tau[2]+Gamma)/(Gamma-Alpha);
	
	// Position coordinates
	x[0] = sqrt(R2)*cos(tau[1]);
	x[1] = sqrt(R2)*sin(tau[1]);
	x[2] = sqrt(z2);
	
	double Rdot = 0.5*((tau[0]+Alpha)*taudot[2]+(tau[2]+Alpha)*taudot[0])/((Alpha-Gamma)*sqrt(R2));
	double zdot = 0.5*((tau[0]+Gamma)*taudot[2]+(tau[2]+Gamma)*taudot[0])/((Gamma-Alpha)*x[2]);
	
	// Velocity coordinates
	v[0] = Rdot*cos(tau[1])-taudot[1]*x[1];
	v[1] = Rdot*sin(tau[1])+taudot[1]*x[0];
	v[2] = zdot;
	
	return;
	}

void CoordSys::tau2Rz(double tau[2], double x[2]){
// Calculates the cylindrical polar coordinates (R,z) given the tau coordinates (lambda,nu)

	double R2 = (tau[0]+Alpha)*(tau[1]+Alpha)/(Alpha-Gamma);
	double z2 = (tau[0]+Gamma)*(tau[1]+Gamma)/(Gamma-Alpha);
	if (R2<0.0) R2=0.0;
	if (z2<0.0) z2=0.0;
	x[0] = sqrt(R2);
	x[1] = sqrt(z2);

	return;
	}

void CoordSys::RzvRvz2tautaudot(double x[4], double tau[4]){
// Calculates tau and taudot (not phi and phidot) given the polar coordinates and velocities
// (R,z,vR,vz)
    
	double b = Alpha+Gamma-x[0]*x[0]-x[1]*x[1];
  	double c = Alpha*Gamma-Gamma*x[0]*x[0]-Alpha*x[1]*x[1];
  	double bdot = -2.0*(x[0]*x[2]+x[1]*x[3]);
  	double cdot = -2.0*(Gamma*x[0]*x[2]+Alpha*x[1]*x[3]);
  	
  	tau[0] = 0.5*(-b+sqrt(b*b-4.0*c)); // lambda
  	tau[1] = 0.5*(-b-sqrt(b*b-4.0*c)); // nu
	tau[2] = 0.5*(-bdot+(b*bdot-2.0*cdot)/sqrt(b*b-4.0*c));
	tau[3] = 0.5*(-bdot-(b*bdot-2.0*cdot)/sqrt(b*b-4.0*c));
	return;
	}
	
void CoordSys::RzvRvz2tautaudot(particle &star){
// As above, but operates on the particle class
    
	double b = Alpha+Gamma-star.Rz[0]*star.Rz[0]-star.Rz[1]*star.Rz[1];
  	double c = Alpha*Gamma-Gamma*star.Rz[0]*star.Rz[0]-Alpha*star.Rz[1]*star.Rz[1];
  	double bdot = -2.0*(star.Rz[0]*star.vRvz[0]+star.Rz[1]*star.vRvz[1]);
  	double cdot = -2.0*(Gamma*star.Rz[0]*star.vRvz[0]+Alpha*star.Rz[1]*star.vRvz[1]);
  	
  	star.tau[0] = 0.5*(-b+sqrt(b*b-4.0*c)); // lambda
  	star.tau[1] = star.phi;
  	star.tau[2] = 0.5*(-b-sqrt(b*b-4.0*c)); // nu
	star.tau[3] = 0.5*(-bdot+(b*bdot-2.0*cdot)/sqrt(b*b-4.0*c));
	star.tau[4] = star.lz/(star.Rz[0]*star.Rz[0]);
	star.tau[5] = 0.5*(-bdot-(b*bdot-2.0*cdot)/sqrt(b*b-4.0*c));

	return;
	}

double CoordSys::I3(double x[4], double Energy, double I2){
// Calculates I3 given a (R,z,vR,vz) point, coordinate system and the two other isolating integrals
// This function needs G, so we must fit the potential before calling this
	double tau[4];
	RzvRvz2tautaudot(x,tau);
	return (tau[0]+Gamma)*(Energy-(I2/(tau[0]+Alpha))+G(tau[0]))-(pow((tau[2]*(tau[0]-tau[1])),2.0))/(8.0*(tau[0]+Alpha)*(tau[0]+Gamma));
	}


void CoordSys::calcIntegrals(particle &star){
///Calculates the three classical integrals H, I2, I3 given cartesian coordinates x and velocities v - must be passed common structure

	RzvRvz2tautaudot(star);
	
  double E =  0.5*(star.vRvz[0]*star.vRvz[0]+star.vRvz[1]*star.vRvz[1]+star.lz*star.lz/(star.Rz[0]*star.Rz[0]))+Phi(star.Rz[0],star.Rz[1]);
  double I2 = 0.5*(pow(star.lz,2.0));
  star.Int[0] = E;
  star.Int[1] = I2;
  
  // We find I3 by calculating an average of I3 estimates for the three edge points (lamplus,lamminus,nuplus)
  star.Int[2] = (I3(star.R1,E,I2)+I3(star.R2,E,I2)+I3(star.z1,E,I2))/3.0;
  
  // However, to ensure we do not exclude the given phase-space point from the orbit region we
  // check whether it has been excluded and if so adjust I3 so it is included
  // lambda check
  double pcheck = E-I2/(star.tau[0]+Alpha)-star.Int[2]/(star.tau[0]+Gamma)+G(star.tau[0]);
  if(pcheck<0.0){
  	double RR[4]={star.Rz[0],star.Rz[1],star.vRvz[0],star.vRvz[1]};
  	star.Int[2] = I3(RR,E,I2);
  }
  // nu check
  pcheck = E-I2/(star.tau[1]+Alpha)-star.Int[2]/(star.tau[1]+Gamma)+G(star.tau[1]);
  if(pcheck>0.0){
  	double RR[4]={star.Rz[0],star.Rz[1],star.vRvz[0],star.vRvz[1]};
  	star.Int[2] = I3(RR,E,I2);
  }

	// Check to ensure orbit crosses zplane - this was put here to make p_tau real on the z=0 axis
	// However, I think this only occurs with ill-conditioned input. Left in just in case.
	// Bit dangerous though as sometimes makes p_lam^2<0 for all lam... Hence the 0.999999
  double pt = E-I2/(1e-10-Gamma+Alpha)-star.Int[2]/(1e-10)+G(1e-10-Gamma);
  if(pt>0.0){
  	while(pt>0.0){
  		star.Int[2]*=0.999999;
  		pt = E-I2/(1e-10-Gamma+Alpha)-star.Int[2]/(1e-10)+G(1e-10-Gamma);
  	} 
  	std::cout << "I3 altered..." << std::endl;
  	star.FLAG=1;
  }

//	std::cout<< E-I2/(10.-Gamma+Alpha)-star.int[2]/(10.)+G(10.-Gamma) << std::endl;
  return;
}

//=================================================================================================

////////////////////////////////
//Particle Class Constructors //
////////////////////////////////

particle::particle(double x1, double x2, double x3, double v1, double v2, double v3){
// creates a particle at Cartesian 6D phase space point (x1,x2,x3,v1,v2,v3) and calculates the 
// corresponding polar coordinates
	x[0] = x1;x[1] = x2;x[2] = x3;
	v[0] = v1;v[1] = v2;v[2] = v3;
	Rz[0] = sqrt(x1*x1+x2*x2);
	Rz[1] = x3;
	phi = atan2(x2,x1);
	vRvz[0] = v1*cos(phi)+v2*sin(phi);
	vRvz[1] = v3;
	lz = x1*v2-x2*v1;
	FLAG=0;
	}

	
particle::particle(double R1, double R2, double vR1, double vphi1, double vz1){
// creates particle at polar point
	Rz[0] = R1;Rz[1] = R2;
	vRvz[0] = vR1;vRvz[1] = vz1;
	lz = R1*vphi1;
	FLAG=0;
}

particle::particle(double xx[],int type){
// does above but on array, if type==0 -> Cartesian, type==1 -> Polar
	if(type==0){ // Cartesian
	x[0] = xx[0];x[1] = xx[1];x[2] = xx[2];
	v[0] = xx[3];v[1] = xx[4];v[2] = xx[5];
	Rz[0] = sqrt(xx[0]*xx[0]+xx[1]*xx[1]);
	Rz[1] = xx[2];
	phi = atan2(xx[1],xx[0]);
	vRvz[0] = xx[3]*cos(phi)+xx[4]*sin(phi);
	vRvz[1] = xx[5];
	lz = xx[0]*xx[4]-xx[1]*xx[3];
	}
	if(type==1){ //Polar
	Rz[0] = xx[0];Rz[1] = xx[1];
	vRvz[0] = xx[2];vRvz[1] = xx[4];
	lz = xx[0]*xx[3];
	}
	FLAG=0;
	}
	
//=================================================================================================
// StackFit class functions - for initialising potential to be fit and calculating AA

StackFit::StackFit(double T, double S, double R){initPhi(T,S,R);}

void StackFit::ActionsAngles(double x[6], double aa[6]){
// this function takes a 6D phase space point x[6] and outputs the actions and angles
// in aa[6]. The potential must be initialised before entry.

  	particle star(x,0);
  	
  	// Fit potential
	fitpot(star);

	// Setup Coordinate System
	CoordSys Stackel (Alpha,Gamma);
    Stackel.calcIntegrals(star);

	Actions(Stackel,star);
	Angles(Stackel,star);

	aa[0] = star.J[0];aa[1] = star.lz;aa[2] = star.J[1];
	aa[3] = star.Angs[0];aa[4] = star.Angs[1];aa[5] = star.Angs[2];
	
    }
    
void StackFit::ActionsAnglesPolar(double x[5], double aa[6]){
// this function takes a 5D phase space point x[5] (R,z,vR,vphi,vz) and outputs the actions and angles
// in aa[6]. The potential must be initialised before entry.

  	particle star(x,1);
  	
  	// Fit potential
	fitpot(star);

	// Setup Coordinate System
	CoordSys Stackel (Alpha,Gamma);
    Stackel.calcIntegrals(star);

	Actions(Stackel,star);
	Angles(Stackel,star);

	aa[0] = star.J[0];aa[1] = star.lz;aa[2] = star.J[1];
	aa[3] = star.Angs[0];aa[4] = star.Angs[1];aa[5] = star.Angs[2];
	
    }

// Action and angle routines //

void StackFit::Actions(CoordSys &Stackel, particle &star){

	IntGL[0]=star.Int[0];IntGL[1]=star.Int[1];IntGL[2]=star.Int[2];

	// Need to find limits
	double lam0,lam1,nu0=-Gamma+TINY,nu1;//mu0,mu1,
	
	//J_lam
	
	// First we must find a value of lambda which is allowed - not necessarily the input lambda
	double tauCopy=star.tau[0];
	if(ptau2ROOT(tauCopy)>0.0){
	double dtau=Stackel.Alpha*0.1;
	tauCopy+=dtau;double plast=0.0, p2=ptau2ROOT(tauCopy);
	while(p2<0.0){
		plast=p2;
		tauCopy+=dtau;
		p2=ptau2ROOT(tauCopy);
		if(p2<plast)dtau*=-0.5;
	}
	}
	lam0=lam1=tauCopy;
	
	// find roots of p^2(lam)
	double laminner=lam0;
	
	// A finer search might be necessary here
	while(ptau2ROOT(laminner)>0.)	laminner-=.01*(lambdaplus-lambdaminus)/ACCURACY;
	double lamouter=lam0;
	while(ptau2ROOT(lamouter)>0.)	lamouter+=.01*(lambdaplus-lambdaminus)/ACCURACY;
	
	lam0=root_brent(&ptau2ROOT,laminner,lam0,ptau2ROOT(laminner),ptau2ROOT(lam0),TINY,100);
	lam1=root_brent(&ptau2ROOT,lam1,lamouter,ptau2ROOT(lam1),ptau2ROOT(lamouter),TINY,100);
	
	// perform J_lam integral
	taubargl = 0.5*(lam0+lam1); Deltagl=0.5*(lam1-lam0);
	star.J[0] = Deltagl*GaussLegendreQuad(&Jint,-.5*PI,.5*PI)/PI;

	star.J[1]=sqrt(2.0*star.Int[1]);

	//J_nu
	nu0=-Stackel.Gamma+TINY;
	// Check to ensure orbit crosses zplane - this was put here to make p_tau real on the z=0 axis
	// However, I think this only occurs with ill-conditioned input. Left in just in case.
	// Bit dangerous though as sometimes makes p_lam^2<0 for all lam... This is why we put it here, 
	// after finding the lam limits
  	double pt = ptau2ROOT(nu0);
  	if(pt>0.0){
  		while(pt>0.0){
  			star.Int[2]*=0.9999999999;
  			IntGL[2]=star.Int[2];
  			pt = ptau2ROOT(nu0);
  		} 
  		//std::cout << "I3 altered..." << std::endl;
  		star.FLAG=1;
  	}
	
	nu1=star.tau[2];
	// find root of p^2(nu)
	double nuouter=1.0*nu0;
	
	// A finer search might be necessary here
	while(ptau2ROOT(nuouter)<0.)	nuouter+=0.01*(nuplus-numinus)/ACCURACY;
	if(nuouter<-Stackel.Alpha)nu1=root_brent(&ptau2ROOT,nu0,nuouter,ptau2ROOT(nu0),ptau2ROOT(nuouter),TINY,100);
	else nu1=-Stackel.Alpha-TINY;

	// perform J_nu integral
	taubargl = 0.5*(nu0+nu1); Deltagl=0.5*(nu1-nu0);
	star.J[2] = 2.0*Deltagl*GaussLegendreQuad(&Jint,-.5*PI,.5*PI)/PI;
	star.limits[0]=lam0;star.limits[1]=lam1;star.limits[2]=nu0;star.limits[3]=nu1;
	
	//Sometimes..., lam0=lam1 -> shell orbit. Want to avoid so set lam1=lam0+SMALL
	if(lam1<=lam0){star.limits[1]=star.limits[0]+SMALL;star.FLAG=1;}
	}

void StackFit::Angles(CoordSys &Stackel, particle &star){

	IntGL[0]=star.Int[0];IntGL[1]=star.Int[1];IntGL[2]=star.Int[2];
	
	// dIdJ ==================================================================
	double dJdI[3][3]; double dIdJ[3][3];
	//Lam
	taubargl = 0.5*(star.limits[0]+star.limits[1]); Deltagl=0.5*(star.limits[1]-star.limits[0]);
	dJdI[0][0] = Deltagl*GaussLegendreQuad(&dJdEint,-.5*PI,.5*PI)/PI;
	dJdI[0][1] = Deltagl*GaussLegendreQuad(&dJdI2int,-.5*PI,.5*PI)/PI;
	dJdI[0][2] = Deltagl*GaussLegendreQuad(&dJdI3int,-.5*PI,.5*PI)/PI;
	
	dJdI[1][0] = 0.0;
	dJdI[1][1] = 1.0/(sqrt(2.0*star.Int[1]));
	dJdI[1][2] = 0.0;	
	
	//Nu
	taubargl = 0.5*(star.limits[2]+star.limits[3]); Deltagl=0.5*(star.limits[3]-star.limits[2]);
	dJdI[2][0] = Deltagl*GaussLegendreQuad(&dJdEint,-.5*PI,.5*PI)/PI;
	dJdI[2][1] = Deltagl*GaussLegendreQuad(&dJdI2int,-.5*PI,.5*PI)/PI;
	dJdI[2][2] = Deltagl*GaussLegendreQuad(&dJdI3int,-.5*PI,.5*PI)/PI;

	double Determinant = dJdI[0][0]*det2(dJdI[1][1],dJdI[1][2],dJdI[2][1],dJdI[2][2])
						- dJdI[0][1]*det2(dJdI[1][0],dJdI[1][2],dJdI[2][0],dJdI[2][2])
						+ dJdI[0][2]*det2(dJdI[1][0],dJdI[1][1],dJdI[2][0],dJdI[2][1]);

	dIdJ[0][0] = det2(dJdI[1][1],dJdI[1][2],dJdI[2][1],dJdI[2][2])/Determinant;
	dIdJ[0][1] = -det2(dJdI[0][1],dJdI[0][2],dJdI[2][1],dJdI[2][2])/Determinant;
	dIdJ[0][2] = det2(dJdI[0][1],dJdI[0][2],dJdI[1][1],dJdI[1][2])/Determinant;
	dIdJ[1][0] = -det2(dJdI[1][0],dJdI[1][2],dJdI[2][0],dJdI[2][2])/Determinant;
	dIdJ[1][1] = det2(dJdI[0][0],dJdI[0][2],dJdI[2][0],dJdI[2][2])/Determinant;
	dIdJ[1][2] = -det2(dJdI[0][0],dJdI[0][2],dJdI[1][0],dJdI[1][2])/Determinant;
	dIdJ[2][0] = det2(dJdI[1][0],dJdI[1][1],dJdI[2][0],dJdI[2][1])/Determinant;
	dIdJ[2][1] = -det2(dJdI[0][0],dJdI[0][1],dJdI[2][0],dJdI[2][1])/Determinant;
	dIdJ[2][2] = det2(dJdI[0][0],dJdI[0][1],dJdI[1][0],dJdI[1][1])/Determinant;

	// dSdI ===================================================================
	double dSdE=0.0,dSdI2=0.0,dSdI3=0.0;
	//Lam
	taubargl = 0.5*(star.limits[0]+star.limits[1]); Deltagl=0.5*(star.limits[1]-star.limits[0]);
	double thetaLimit = asin(MAX(-1.0,MIN(1.0,(star.tau[0]-taubargl)/Deltagl)));
	double lsign = SIGN(star.tau[3]);

	dSdE  += lsign*Deltagl*GaussLegendreQuad(&dJdEint,-.5*PI,thetaLimit);
	dSdI2 += lsign*Deltagl*GaussLegendreQuad(&dJdI2int,-.5*PI,thetaLimit);
	dSdI3 += lsign*Deltagl*GaussLegendreQuad(&dJdI3int,-.5*PI,thetaLimit);
	
	dSdI2+=SIGN(star.tau[4])*star.tau[1]/(sqrt(2.0*star.Int[1]));
	
	//Nu
	taubargl = 0.5*(star.limits[2]+star.limits[3]); Deltagl=0.5*(star.limits[3]-star.limits[2]);
	thetaLimit = asin(MAX(-1.0,MIN(1.0,(star.tau[2]-taubargl)/Deltagl)));
	double nsign = SIGN(star.tau[5]);

	dSdE  += nsign*Deltagl*GaussLegendreQuad(&dJdEint,-.5*PI,thetaLimit);
	dSdI2 += nsign*Deltagl*GaussLegendreQuad(&dJdI2int,-.5*PI,thetaLimit);
	dSdI3 += nsign*Deltagl*GaussLegendreQuad(&dJdI3int,-.5*PI,thetaLimit);
	
	star.Angs[0] = dSdE*dIdJ[0][0]+dSdI2*dIdJ[1][0]+dSdI3*dIdJ[2][0];
	star.Angs[1] = dSdE*dIdJ[0][1]+dSdI2*dIdJ[1][1]+dSdI3*dIdJ[2][1];
	star.Angs[2] = dSdE*dIdJ[0][2]+dSdI2*dIdJ[1][2]+dSdI3*dIdJ[2][2];

	if(star.Angs[0]<0.0){star.Angs[0]+=2.0*PI;}
	if(star.Angs[1]<0.0){star.Angs[1]+=2.0*PI;}
	if(star.tau[5]<0.0){star.Angs[2]+=2.0*PI;}
	if(star.Rz[1]<0.0){star.Angs[2]+=2.0*PI;}
	star.Angs[2]/=2.0;
	if(star.Angs[2]>2.*PI)star.Angs[2]-=2.*PI;
	if(star.Angs[2]<0.0)star.Angs[2]+=2.*PI;
	
	Determinant = dJdI[0][0]*det2(dJdI[1][1],dJdI[1][2],2.0*dJdI[2][1],2.0*dJdI[2][2])
				- dJdI[0][1]*det2(dJdI[1][0],dJdI[1][2],2.0*dJdI[2][0],2.0*dJdI[2][2])
				+ dJdI[0][2]*det2(dJdI[1][0],dJdI[1][1],2.0*dJdI[2][0],2.0*dJdI[2][1]);

	star.freq[0]=det2(dJdI[1][1],dJdI[1][2],2.0*dJdI[2][1],2.0*dJdI[2][2])/Determinant;
	star.freq[1]=det2(2.0*dJdI[2][1],2.0*dJdI[2][2],dJdI[0][1],dJdI[0][2])/Determinant;
	star.freq[2]=det2(dJdI[0][1],dJdI[0][2],dJdI[1][1],dJdI[1][2])/Determinant;

	// FLAG data which doesn't make sense

	if(star.freq[0]<0.0 || star.freq[1]<0.0 || star.freq[2]<0.0 ||
		star.Angs[0]<0.0 || star.Angs[0]>2.*PI ||
		star.Angs[1]<0.0 || star.Angs[1]>2.*PI ||
		star.Angs[2]<0.0 || star.Angs[2]>2.*PI ||
		star.Angs[0]!=star.Angs[0] ||
		star.Angs[1]!=star.Angs[1] ||
		star.Angs[2]!=star.Angs[2] ) star.FLAG=1;
	}
//=================================================================================================

///////////////////////
// Printing Functions//
///////////////////////

void FilePrint(std::ofstream & OUT, particle &star){
	OUT << std::setprecision(9) << std::scientific << star.J[0] << " " << star.lz << " " << star.J[2] << " " << star.Angs[0] 
	<< " " << star.Angs[1] << " " << star.Angs[2] << " " << star.Int[0] << " " << star.Int[1] 
	<< " " << star.Int[2] << " " << star.x[0] << " " << star.x[1] << " " << star.x[2] << " " 
	<< star.v[0] << " " << star.v[1] << " " << star.v[2] << " " << Alpha << " " << star.R1[0] << " " 
	<< star.R1[1] << " " << star.R2[0] << " " << star.R2[1] << " " << star.z1[0] << " " << star.z1[1] << " " << star.FLAG << " "
	<< (std::abs(VFit(star.Rz[0],star.Rz[1])-Phi(star.Rz[0],star.Rz[1]))/(std::abs(Phi(star.Rz[0],star.Rz[1])))) 
	<<" "<<star.freq[0]<<" "<<star.freq[1]<<" "<<star.freq[2]<<std::endl;
	}

void nicePrint(float j[2], double I2, double ang[3], double freq[3]){
// prints actions, angles and frequencies in nice way
	double Lz = sqrt(2.0*I2);
	printf("\nActions and Angles\n%-15.9g%-15.9g%-15.9g%-15.9g%-15.9g%-15.9g\n\nFrequencies\n%-15.9g%-15.9g%-15.9g\n\n",
				j[0],Lz,j[1],ang[0],ang[1],ang[2],freq[0],freq[1],freq[2]);
	return;
}

void printHess(double hess[3][3]){
// prints Hess matrix in matrixy way :)
	printf("Hessian\n");
	for (int i=0;i<3;i++){
  		for (int j=0;j<3;j++){
			printf("%-15.6g",hess[i][j]);}
		printf("\n");
		}
	printf("\n");
	return;
}

//=================================================================================================
