/*

fitpot.cpp

Jason Sanders, jason.sanders@physics.ox.ac.uk

contains code to find the limits of the orbit and then find a good guess of Alpha for this region

*/
//=================================================================================================

#include <cmath>
#include <stdio.h>
#include "PotMN.h"
#include "fitpot.h"
#include "aa.h"
#include "GFn.h"
#include "utils.h"

//=================================================================================================

// Globals for performing fit integrals
double lambdaminus,lambdaplus,numinus,nuplus;
double L,N,chibar,nuIntTmp,lamIntTmp;
double zsign;

// Coordinate System Definitions
double Alpha; double Gamma = -1.0;

// Ang Mom for orbit integrator
double LZ;

// Number of data points for interpolation table
int DATAPOINTS = 40;

//=================================================================================================
void FreeMem(){
	delete lambdagrid;delete nugrid;delete flamgrid;
	delete fnugrid; delete y2Lamgrid; delete y2Nugrid;
}
void Rz2tau(double x[], double tau[]){
  // Calculates the prolate spheroidal coordinates tau = (lambda,phi,nu) given cartesian coordinates x = (x,y,z)
    
  double R2 = x[0]*x[0];

  double b = Alpha+Gamma-R2-x[1]*x[1];
  double c = Alpha*Gamma-Gamma*R2-Alpha*x[1]*x[1];

  tau[0] = 0.5*(-b+sqrt(b*b-4.0*c)); // lambda
  tau[1] = 0.5*(-b-sqrt(b*b-4.0*c)); // nu

  return;
      }

void tau2Rz(double tau[], double x[2]){
// Calculates the cylindrical polar coordinates (R,z) given the tau coordinates (lambda,nu)

	double R2 = (tau[0]+Alpha)*(tau[1]+Alpha)/(Alpha-Gamma);
	double z2 = (tau[0]+Gamma)*(tau[1]+Gamma)/(Gamma-Alpha);
	if (R2<0.0) R2=0.0;
	if (z2<0.0) z2=0.0;
	x[0] = sqrt(R2);
	x[1] = sqrt(z2);

	return;
	}

//=================================================================================================

/////////////////////////////////
// A LOAD OF FUNCTIONS FOR FIT //
/////////////////////////////////
	
double Vlamnu(double lambda, double nu){
// potential to be fitted to as a function of tau coordinates
	double tau[2] = {lambda,nu};
	double x[2];
	tau2Rz(tau,x);
	return Phi(x[0],x[1]);
	}

double chi(double lambda, double nu){
	double V = Vlamnu(lambda,nu);
	return -(lambda-nu)*V;
	}

double BigNu(double nu){
//Nu weight - see Paper 1 with d=0
	double cv = 0.0;
	return pow((nu+Gamma)/(Gamma-Alpha),cv);
	}
	
double BigLambda(double lambda){
//Lambda weight - see Paper 1 with d=0
	double cl = 5.0;
	return pow((1.0/lambda),cl);
	}

double chibarint(double lambda,double nu){
// integrand for \bar{\chi}
	return chi(lambda,nu)*BigLambda(lambda)*BigNu(nu);
	}

void ChiBarIntCubature(unsigned ndim, const double *x, void *fdata, unsigned fdim, double *fval) {
// interface to integrand for chibar for use with cubature code
    fval[0] = chibarint(x[0],x[1]);
}

double chiNu(double nu){
	return chi(lamIntTmp,nu)*BigNu(nu);
	}
	
double chiLam(double lambda){
	return chi(lambda,nuIntTmp)*BigLambda(lambda);
	}

double flam(double lambda){
	lamIntTmp=lambda;
	return ((GaussLegendreQuad(&chiNu,numinus,nuplus)/N)-0.5*chibar);
}
	
double fnu(double nu){
	nuIntTmp=nu;
	return (-(GaussLegendreQuad(&chiLam,lambdaminus,lambdaplus)/L)+0.5*chibar);
}
	
double VfitLamNu(double lambda, double nu){
	return -(flamINT(lambda)-fnuINT(nu))/(lambda-nu);
	}

double VFit(double R, double z){
	double tau[2];
	double x[2] = {R,z};
	Rz2tau(x,tau);
	return VfitLamNu(tau[0],tau[1]);
	}

//=================================================================================================
// Method for guessing Alpha given an (R,z) point using the second derivatives of the potential

// Second derivs - all numerical finite difference
double d2PhidR2(double R,double z){
	double dPhidRplus, dPhidRminus;
	double tmp;
	double P = Phi(R,z);
	dPhi(R+0.001,z,dPhidRplus,tmp);
	dPhi(R-0.001,z,dPhidRminus,tmp);
	return (dPhidRplus-dPhidRminus)/0.002;
	}
			
double d2Phidz2(double R,double z){
	double dPhidzplus,dPhidzminus;
	double tmp;
	double P = Phi(R,z);
	dPhi(R,z+0.001,tmp,dPhidzplus);
	dPhi(R,z-0.001,tmp,dPhidzminus);
	return (dPhidzplus-dPhidzminus)/0.002;
	}
	
double d2PhidzdR(double R,double z){
	double dPhidzplus,dPhidzminus;
	double tmp;
	double P = Phi(R,z);
	dPhi(R+0.001,z,tmp,dPhidzplus);
	dPhi(R-0.001,z,tmp,dPhidzminus);
	return (dPhidzplus-dPhidzminus)/0.002;
	}

double AlphaGuess(double x[]){
	// Returns a guess of Alpha assuming dldv((l-v)V)=0
	double R = x[0];
	double z = x[1];
	double dR, dz;
	dPhi(R,z,dR,dz);
	double d2dR2 = d2PhidR2(R,z);
	double d2dz2 = d2Phidz2(R,z);
	double d2dzdR = d2PhidzdR(R,z);
	double alphareturn = -fabs(z*z-R*R+(3.0*z*dR-3.0*R*dz+R*z*(d2dR2-d2dz2))/d2dzdR)+Gamma;
	if(alphareturn!=alphareturn){
	// To avoid NaN
	double dxdt[4];derivs(0.0,x,dxdt);ode_rungekutta4(x,dxdt,4,0.0,0.5,&derivs);alphareturn = AlphaGuess(x);
	}
	return alphareturn;
	}

//=================================================================================================

void RzvRvz2tautaudot(double x[4], double tau[4],double alpha){
    // converts polar coordinates and velocities to tau and taudot
    // this version differs from that in CoordSys class as this one depends on the coordinate system
    // so can be used without creating a coordinate system
    
	double b = alpha+Gamma-x[0]*x[0]-x[1]*x[1];
  	double c = alpha*Gamma-Gamma*x[0]*x[0]-alpha*x[1]*x[1];
  	double bdot = -2.0*(x[0]*x[2]+x[1]*x[3]);
  	double cdot = -2.0*(Gamma*x[0]*x[2]+alpha*x[1]*x[3]);
  	
  	tau[0] = 0.5*(-b+sqrt(b*b-4.0*c)); // lambda
  	tau[1] = 0.5*(-b-sqrt(b*b-4.0*c)); // nu
	tau[2] = 0.5*(-bdot+(b*bdot-2.0*cdot)/sqrt(b*b-4.0*c));
	tau[3] = 0.5*(-bdot-(b*bdot-2.0*cdot)/sqrt(b*b-4.0*c));
	return;
	}


void derivs(double x, double y[], double dydx[]){
	// RHS for equations of motion for orbit integration
	
	dydx[0] = y[2]; //vR
	dydx[1] = y[3]; //vz
	
	double dVdR, dVdz;
	dPhi(y[0],y[1],dVdR,dVdz); //find derivs of potential
	
	//equations of motion
	dydx[2] = (LZ*LZ/(y[0]*y[0]*y[0]))-dVdR; 
	dydx[3] = -dVdz;
	
	}

void fitpot(particle &star){
	// This function begins by integrating the initial phase space point with an adaptive timestep
	// to an accuracy of 0.01% and finding an estimate of alpha using the first few points.
	// We then determine the three points which define the fitting region - lambdaminus, lambdaplus 
	// and nuplus which are defined by taudot=0.
	// We perform the potential fit over the region defined by these three points.
	// At the end, we fill the arrays to be used in the interpolation of the f functions.	
	
	double y[4] = {star.Rz[0],star.Rz[1],star.vRvz[0],star.vRvz[1]}; // initial conditions
	double ytmp[4]; // for storing initial point of integration at each step
	double t = 0.0; // D.E is time indep so this can be anything
	double dydx[4]; // derivative array
	double yscal[4] = {1.0,1.0,1.0,1.0}; // used for weighting the relative importance of errors in each variable
	double StepAchieved, NextStep = 0.5; // for adaptive timestepping
	double tautmp[4]; // to store tau coordinates
	int steps=0; // counts no. of int. steps
	LZ = star.lz;
		
	// flags for checking when three edge points are found
	int alldone1=0;
	int alldone2=0;
	int alldone3=0;
	
	double alphatmp=AlphaGuess(y); // Initial guess of alpha
	int n = 2; // counts no. of alpha guesses+1

	double nutmp,lamtmp; // stores previous lambda and nu values to determine taudot.

	int NMAX = 6; // Number of bisections to perform
	int STEPMAX=50;
	
	double ORBACCURACY = 0.000001/ACCURACY;
	
	while (alldone1*alldone2*alldone3==0){

		// Perform one time step with adaptive timestep and accuracy 0.01%
		derivs(t,y,dydx);
		ytmp[0]=y[0];ytmp[1]=y[1];ytmp[2]=y[2];ytmp[3]=y[3];
		ode_rungekuttaAdaptiveStep(y,dydx,4,&t,NextStep,ORBACCURACY,yscal,&StepAchieved,&NextStep,&derivs);
		
		// Refine our guess of alpha
		if(steps%5==0 && steps<=STEPMAX){
			alphatmp = ((n-1)*alphatmp+AlphaGuess(y))/(double)n;n++;
			}
			
		RzvRvz2tautaudot(y,tautmp,alphatmp);
		double nudotdot = (tautmp[3]-nutmp);
		double lamdotdot = (tautmp[2]-lamtmp);
		
		// Edge is defined by points when lambdadot or nudot change sign

		// Inner Lambda Edge
		if(tautmp[2]*lamtmp<0.0 && lamdotdot>0.0 && steps>10 && alldone1==0){
		
			// Now we know the interval the root occurs in we can find the root by repeated bisection
			
			double STEP = StepAchieved/2.0;
			int N = 0;
			double ylower[4];ylower[0]=ytmp[0];ylower[1]=ytmp[1];ylower[2]=ytmp[2];ylower[3]=ytmp[3];
			y[0]=ytmp[0];y[1]=ytmp[1];y[2]=ytmp[2];y[3]=ytmp[3];
			double tauLoop[4];
			double lamLoop = lamtmp;
			
			// Carry out N bisections
			while(N<NMAX){
				
				// Carry out integration over half the interval
				derivs(t,y,dydx);
				ode_rungekutta4(y,dydx,4,t,STEP,&derivs);
				RzvRvz2tautaudot(y,tauLoop,alphatmp);
				
				// Check which half root occurs in
				if(tauLoop[2]*lamLoop<0.0){
					y[0]=ylower[0];y[1]=ylower[1];y[2]=ylower[2];y[3]=ylower[3];
					}
				else{
					ylower[0]=y[0];ylower[1]=y[1];ylower[2]=y[2];ylower[3]=y[3];lamtmp=tauLoop[2];
					}
				
				STEP = STEP/2.0;
				N++;
			}
			star.R1[0]=y[0];star.R1[1]=y[1];star.R1[2]=y[2];star.R1[3]=y[3];lambdaminus = tauLoop[0];alldone1=1;
			}
		
		// Outer Lambda Edge - same algorithm as above
		if(tautmp[2]*lamtmp<0.0 && lamdotdot<0.0 && steps>10 && alldone2==0){
			
			double STEP = StepAchieved/2.0;
			int N = 0;
			double ylower[4];ylower[0]=ytmp[0];ylower[1]=ytmp[1];ylower[2]=ytmp[2];ylower[3]=ytmp[3];
			y[0]=ytmp[0];y[1]=ytmp[1];y[2]=ytmp[2];y[3]=ytmp[3];
			double tauLoop[4];
			while(N<NMAX){
				
				derivs(t,y,dydx);
				ode_rungekutta4(y,dydx,4,t,STEP,&derivs);
				RzvRvz2tautaudot(y,tauLoop,alphatmp);
				
				if(tauLoop[2]*lamtmp<0.0){
					y[0]=ylower[0];y[1]=ylower[1];y[2]=ylower[2];y[3]=ylower[3];
					}
				else{
					ylower[0]=y[0];ylower[1]=y[1];ylower[2]=y[2];ylower[3]=y[3];lamtmp=tauLoop[2];
					}
				STEP = STEP/2.0;
				N++;
			}

			star.R2[0]=y[0];star.R2[1]=y[1];star.R2[2]=y[2];star.R2[3]=y[3];lambdaplus = tauLoop[0];alldone2=1;
		}
	
		// Upper Nu Edge - same algorithm as above
		if(nutmp*tautmp[3]<0.0 && nudotdot<0.0 && steps>10 && alldone3==0){

			double STEP = StepAchieved/2.0;
			int N = 0;
			double ylower[4];ylower[0]=ytmp[0];ylower[1]=ytmp[1];ylower[2]=ytmp[2];ylower[3]=ytmp[3];
			y[0]=ytmp[0];y[1]=ytmp[1];y[2]=ytmp[2];y[3]=ytmp[3];
			double tauLoop[4];
			while(N<NMAX){
				
				derivs(t,y,dydx);
				ode_rungekutta4(y,dydx,4,t,STEP,&derivs);
				RzvRvz2tautaudot(y,tauLoop,alphatmp);
				nudotdot = (tauLoop[3]-nutmp);

				if(tauLoop[3]*nutmp<0.0 && nudotdot<0.0){
					y[0]=ylower[0];y[1]=ylower[1];y[2]=ylower[2];y[3]=ylower[3];
					}
				else{
					ylower[0]=y[0];ylower[1]=y[1];ylower[2]=y[2];ylower[3]=y[3];nutmp=tauLoop[3];
					}
				STEP = STEP/2.0;
				N++;
			}

			star.z1[0]=y[0];star.z1[1]=y[1];star.z1[2]=y[2];star.z1[3]=y[3];nuplus = tauLoop[1];alldone3=1;
		}
		
		lamtmp=tautmp[2];
		nutmp=tautmp[3];
		steps++;
	}

	Alpha = alphatmp;
	numinus = -Gamma;

	// Fit //
	// Perform integrals which are indep. of R,z
	// can do integrals analytically rather than numerically
	L = 0.25*(pow(lambdaminus,-4)-pow(lambdaplus,-4));N = (nuplus-numinus);
	
	//double xmin[2] = {lambdaminus,numinus}, xmax[2] = {lambdaplus,nuplus}, val, err;
   	//adapt_integrate(1, ChiBarIntCubature, NULL, 2, xmin, xmax, 0, 0, 1e-6, &val, &err);
   	double val = DoubleGaussLegendreQuad(&chibarint,lambdaminus,lambdaplus,numinus,nuplus);
	chibar = val/(L*N);

	// Grids for interpolation
	lambdagrid=dmatrix(DATAPOINTS);
	nugrid=dmatrix(DATAPOINTS);
	flamgrid=dmatrix(DATAPOINTS);
	fnugrid=dmatrix(DATAPOINTS);

	// Fill grids
  	for(int i=0;i<DATAPOINTS;i++){
  		lambdagrid[i] = lambdaminus+(double)i*(lambdaplus-lambdaminus)/(double)(DATAPOINTS-1);
		nugrid[i] = numinus+(double)i*(nuplus-numinus)/(double)(DATAPOINTS-1);
		flamgrid[i] = flam(lambdagrid[i]);
		fnugrid[i] = fnu(nugrid[i]);
		}

	// Create splines for interpolation
	y2Lamgrid = dmatrix(DATAPOINTS);
	y2Nugrid = dmatrix(DATAPOINTS);
	interp_spline(lambdagrid, flamgrid, DATAPOINTS, 1e30, 1e30, y2Lamgrid);
	interp_spline(nugrid, fnugrid, DATAPOINTS, 1e30, 1e30, y2Nugrid);

}

//=================================================================================================