/* 

Cartesian to Action-Angle in Stackel potential

xv2aa.cc

Jason Sanders, jason.sanders@physics.ox.ac.uk

Two filenames given in command line - first contains phase space points, 
	second is output file for actions and angles.
1. Initially we find limits of orbit to find the region to fit the potential over.
2. Find a value of Alpha which gives a reasonable fit over the region.\
3. Fit potential over this region by a Stackel potential.
4. Calculate the actions and angles in this Stackel potential.

The required accuracy level can be specified as a third command line argument. This sets the
accuracy of the orbit integration (in fitpot.cpp) as well as the accuracy of the root-finding
 calculations (in aa.cpp). The default ACCURACY = 1.
If you are trying to calculate actions of very radial orbits or the output is dodgy 
(e.g. angles=NAN) then the accuracy should be increased (e.g. ACCURACY=10,100,1000).
 This obviously makes the code run slower. The Miyamoto-Nagai example file works fine 
 without this option being specified (ACCURACY=1). 

Output file:
JR Lz Jz ThetaR ThetaPhi Thetaz E I2 I3 x y z vx vy vz Alpha InnerPointR InnerPointz OuterPointR 
OuterPointz UpperPointR UpperPointz FLAG (Phi_Fit-Phi)/Phi*100_at_initial_point Omega_R Omega_phi 
Omega_z

*/

//=================================================================================================

#include <cmath>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <stdlib.h>
#include "GFn.h"
#include "PotMN.h"
#include "fitpot.h"
#include "aa.h"
#include "utils.h"

//=================================================================================================

int main(int argc, char* argv[]){

	// Initialise a Miyamoto-Nagai potential with GM=1.,a=1.,b=0.45
	StackFit SF(1.,1.0,0.45);

   if (argc > 2) {

    // Read data in from file and output to file
    std::ifstream inFile;
    inFile.open(argv[1],std::ios::in);
    std::ofstream outFile;
    outFile.open(argv[2],std::ios::out);

	if(argc>3){ACCURACY=atoi(argv[3]);}

    if (!inFile.is_open()) std::cout << "Failed to open " << argv[2] << std::endl;
    else{
    double xtmp[6];
    
    while(inFile >> xtmp[0] >> xtmp[1] >> xtmp[2] >> xtmp[3] >> xtmp[4] >> xtmp[5]){
    
		// Input data into star structure
  		particle star(xtmp,0);
  	
  		// Fit potential
		fitpot(star);

		// Setup Coordinate System
		CoordSys Stackel (Alpha,Gamma);
    	Stackel.calcIntegrals(star);
    	
		// Find actions and angles
		SF.Actions(Stackel,star);
		SF.Angles(Stackel,star);

		// Print results to file
		FilePrint(outFile,star);
		
		//Delete Interpolation Grids
		FreeMem();
    }
  
    inFile.close();
	outFile.close();
  	}
}

// If incorrect input has been given 
  else std::cout<<"usage: "<< argv[0] <<" <infilename> <outfilename>\n" << std::endl;
  return 0;
  }

//=================================================================================================